/*
 *      IP nat module definitions
 *      NAT for Linux by Michael Hasenstein
 *		mha@informatik.tu-chemnitz.de
 *		mha@debis-sfi.de
 *	(see RFC 1631)
 *
 *
 *      Copyright (c) 1997 by Michael Hasenstein
 *
 *      --------------------------------------------------------------------
 *	parts of this code, especially the packet matching stuff, have been
 *	taken from linux/net/ipv4/ip_fw.c
 *	see this file for copyrights for this code
 *      --------------------------------------------------------------------
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _IP_NAT_MODULE_H
#define _IP_NAT_MODULE_H

#include <asm/errno.h>
#include <asm/byteorder.h>
#include <net/ip_nat.h>
#include <netinet/in.h>
#include <linux/timer.h>

/* defines for the flag field in struct ip_nat */
#define IP_NAT_DIRECTION_MASK 0x00000003
#define IP_NAT_IN            0x00000001  /* equals IN_NAT in include/net/ip_nat.h */
#define IP_NAT_OUT           0x00000002  /* equals OUT_NAT */
#define IP_NAT_BIDIR         0x00000004
#define IP_NAT_SKIP          0x00000008
#define IP_NAT_LOG           0x00000010
#define IP_NAT_SRNG          0x00000020
#define IP_NAT_DRNG          0x00000040
#define IP_NAT_DYNAMIC       0x00000080
#define IP_NAT_VIRT_SERVER   0x00000100
/* algorithm used for virtual server */ 
#define IP_NAT_VS_RORO       0x00010000  /* round robin */
#define IP_NAT_VS_BYTE       0x00020000  /* total number of bytes transferred */
#define IP_NAT_VS_CONN       0x00040000  /* number of simultaneous connections */

/* commands */
#define IP_NAT_CMD_FLUSH     0x01
#define IP_NAT_CMD_INSERT    0x02
#define IP_NAT_CMD_APPEND    0x04
#define IP_NAT_CMD_DELETE    0x08

/* default timeout */
#define IP_NAT_EXPIRE        10*60*HZ   /* 10 minutes should be plenty */

#define print_ip(a)   printk("%ld.%ld.%ld.%ld",     \
                              (ntohl(a)>>24) & 0xFF,\
                              (ntohl(a)>>16) & 0xFF,\
                              (ntohl(a)>>8)  & 0xFF,\
                              (ntohl(a))     & 0xFF);

union dynamic {
	struct virt_serv *vs;
	struct dyn_nat *dn;
};


#define IP_NAT_MAX_PORTS 10
struct ip_nat {
	struct ip_nat *next;                     /* pointer to next rule */
	union dynamic dynamic;	                 /* pointer to dynamic rules */
	int           id;                        /* unique number for this rule */
        __u32         src, dst;                  /* source/destination ip */
        __u32         src_msk, dst_msk;          /* network masks */
        u_short       nsp, ndp;                  /* number of source/destination ports */
        __u16         ports[IP_NAT_MAX_PORTS];   /* array of port numbers */
	__u8          protocol;                  /* protocol id (udp,tcp,icmp) */
        __u32         via;                       /* interface ip */
	struct device *via_dev;                  /* device of interface "via" */
        __u32         src_nat, dst_nat;          /* new source/destination ip */
        __u32         src_nat_msk, dst_nat_msk;  /* network masks */
	__u16         src_nat_port,dst_nat_port; /* new source/destination port */
	                                         /* Note: dst_nat_port contains the algorithm id */
					         /* instead when rule is a virtual server rule */
	                                         /* virtual servers: algorithm used for load distribution */
	int           timeout;                   /* timeout for some subrules */
	unsigned long pcnt,bcnt;                 /* Packet and byte counters */
        __u32         flag;                      /* flag: bidirectional, skip, packet logging, \
	                                                  direction (in/out), dynamic rule, 
	                                                  virtual server algorithm */
};


/* for calls to setsockopt() (e.g. ipnatadm) */
struct ip_nat_set {
	struct ip_nat  nat;
        char vianame[IFNAMSIZ];  /* interface name */
	int command;
	int weight;              /* for virtual server sub-rules */
	int timeout;
};



/****************************************************************************
 *                         for virtual server rules                         *
 ****************************************************************************/

/* 
 * once per virtual server rule, points to list of:
 *   - real servers
 *   - client connections to servers
 * (is pointed to by "void *ip_nat.dynamic")
 */
struct virt_serv {
	struct server *sr;
	struct client *cl;
	struct server *last;    /* where was the last connection remapped to? */
			        /* used for round-robin algorithm */
};

/* contains info about a real server */
struct server {
	struct server *next;     /* next real server for the virtual server */
	__u32  ip;               /* ip to remap connections to */
	__u16  port;             /* port to remap connections to (can be omitted) */
	unsigned long pcnt,bcnt; /* total number of packets/bytes transferred from/to this server*/
	int connections;         /* number of active connections (TCP only) */
	int weight;              /* some real servers are bigger than others */
};

/* contains info about active client connections to real servers */
struct client {
	struct client *next;      /* to form a list; quite inefficient, should use hashs or trees */
	struct client *prev;
	struct ip_nat *rule;      /* which rule do I belong to? needed for expire-function */
	struct server *server;    /* server the client gets connected to */
	struct timer_list timer;  /* expiration timer */
	__u32  ip;
	__u16  port;
	__u8   protocol;
};

/****************************************************************************/

extern int register_ip_nat_app(void);
extern int unregister_ip_nat_app(void);

#endif /* _IP_NAT_MODULE_H */
