/*
 *      ipnatadm -- IP NAT administration
 *      Version: 97/06/27
 *
 *      See the accompanying manual page ipnatadm(8) for information
 *      about proper usage of this program.
 *
 *	Copyright (c) 1997 by Michael Hasenstein
 *
 *	--------------------------------------------------------------------
 *	large parts of ipnatadm have been taken from ipfwadm, for
 *	this code the following copyright must not be forgotten:
 *
 *	      Copyright (c) 1995,1996 by X/OS Experts in Open Systems BV.
 *      	All rights reserved.
 *
 *	      Author: Jos Vos <jos@xos.nl>
 *
 *              X/OS Experts in Open Systems BV
 *              Kruislaan 419
 *              NL-1098 VA  Amsterdam
 *              The Netherlands
 *
 *              E-mail: info@xos.nl
 *              WWW:    http://www.xos.nl/
 *	--------------------------------------------------------------------
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <getopt.h>
#include "ipnatadm.h"

char package_version[] = "ipnatadm 0.6, 1997/06/27";

struct ip_nat_set nat_set;
char *sports[IP_NAT_MAX_PORTS];
char *dports[IP_NAT_MAX_PORTS];
char *shostnetworkmask, *dhostnetworkmask, *new_shostnetworkmask, *new_dhostnetworkmask;
long options = OPT_NONE;
int naddrs;
struct in_addr *addrs;
int command = CMD_NONE;


int main(int argc, char *argv[])
{
	int c, ret = 0;

	/* set defaults */
	nat_set.nat.id           = 0;
	nat_set.nat.src          = 0;
	nat_set.nat.src_msk      = 0;
	nat_set.nat.dst          = 0;
	nat_set.nat.dst_msk      = 0;
	nat_set.nat.nsp          = 0;
	nat_set.nat.ndp          = 0;
	nat_set.nat.ports[0]     = 0;
	nat_set.nat.protocol     = 0;
	nat_set.nat.src_nat      = 0;
	nat_set.nat.src_nat_msk  = 0;
	nat_set.nat.dst_nat      = 0;
	nat_set.nat.dst_nat_msk  = 0;
	nat_set.nat.src_nat_port = 0;
	nat_set.nat.dst_nat_port = 0;
	nat_set.nat.via          = 0;
	nat_set.nat.timeout      = 0;
	nat_set.nat.flag         = 0;
	nat_set.vianame[0]       = 0;

	while ((c = getopt(argc, argv, "IOFdialht:P:S:D:M:N:V:W:Y:X:w:bsnov")) != -1)
		switch (c) {
			case 'I':
				check_option(OPT_IN, 'I');
				nat_set.nat.flag |= IP_NAT_IN;
				break;
			case 'O':
				check_option(OPT_OUT, 'O');
				nat_set.nat.flag |= IP_NAT_OUT;
				break;
			case 'F':
				if (argc == 2) {
					command = CMD_FLUSH;
				} else
					exit_error(2, "wrong number of arguments (-F can only be used alone)");
				break;
			case 'd':
				if (command != CMD_NONE)
					exit_error(2, "multiple commands specified");
				command = CMD_DELETE;
				break;
			case 'i':
				if (command != CMD_NONE)
					exit_error(2, "multiple commands specified");
				command = CMD_INSERT;
				break;
			case 'a':
				if (command != CMD_NONE)
					exit_error(2, "multiple commands specified");
				command = CMD_APPEND;
				break;
			case 'l':
				if (command != CMD_NONE)
					exit_error(2, "multiple commands specified");
				command = CMD_LIST;
				break;
			case 'h':
				exit_printhelp();
				break;
			case 't':
				check_option(OPT_TIMEOUT, 't');
				nat_set.nat.timeout = HZ * parse_timeout(optarg);
				break;
			case 'P':
				check_option(OPT_PROTOCOL, 'P');
				parse_protocol(optarg);
				break;
			case 'S':
				check_option(OPT_SOURCE, 'S');
				shostnetworkmask = optarg;
				while (optind < argc && argv[optind][0] != '-')
					store_port(argv[optind++], &nat_set.nat.nsp, IP_NAT_SRNG, sports);
				break;
			case 'D':
				check_option(OPT_DESTINATION, 'D');
				dhostnetworkmask = optarg;
				while (optind < argc && argv[optind][0] != '-')
					store_port(argv[optind++], &nat_set.nat.ndp, IP_NAT_DRNG, dports);
				break;
			case 'M':
				check_option(OPT_SRC_NATADDR, 'M');
				new_shostnetworkmask = optarg;
				if (optind < argc && argv[optind][0] != '-')
					nat_set.nat.src_nat_port = parse_port(argv[optind++]);
				break;
			case 'N':
				check_option(OPT_DST_NATADDR, 'N');
				new_dhostnetworkmask = optarg;
				if (optind < argc && argv[optind][0] != '-')
					nat_set.nat.dst_nat_port = parse_port(argv[optind++]);
				break;
			case 'V':
				check_option(OPT_VIAADDR, 'V');
				parse_viahost(optarg);
				break;
			case 'W':
				check_option(OPT_VIANAME, 'W');
				strncpy(nat_set.vianame, optarg, IFNAMSIZ);
				break;
			case 'Y':
				check_option(OPT_ID, 'Y');
				if ((nat_set.nat.id = string_to_number(optarg, 0, INT_MAX-1)) == -1) {
					fprintf(stderr, "ipnatadm: invalid id \"%s\" (too big?)\n", optarg);
					exit_tryhelp(2);
				}
				break;
			case 'X':
				check_option(OPT_VIRTUAL_SERVER, 'X');
				nat_set.nat.flag |= IP_NAT_VIRT_SERVER;
				parse_vs_algorithm(optarg);
				break;
			case 'w':
				check_option(OPT_WEIGHT, 'w');
				if ((nat_set.weight = string_to_number(optarg, 0, INT_MAX-1)) == -1) {
					fprintf(stderr, "ipnatadm: invalid weight \"%s\" (too big?)\n", optarg);
					exit_tryhelp(2);
				}
				break;
			case 'b':
				check_option(OPT_BIDIR, 'b');
				nat_set.nat.flag |= IP_NAT_BIDIR;
				break;
			case 's':
				check_option(OPT_SKIP, 's');
				nat_set.nat.flag |= IP_NAT_SKIP;
				break;
			case 'n':
				check_option(OPT_NUMERIC, 'n');
				break;
			case 'o':
				check_option(OPT_PRINTK, 'o');
				nat_set.nat.flag |= IP_NAT_LOG;
				break;
			case 'v':
				check_option(OPT_VERBOSE, 'v');
				break;

			case '?':
			default:
				exit_tryhelp(2);
		}

	if (optind < argc)
		exit_error(2, "unknown arguments found on commandline");
	else if (!command)
		exit_error(2, "no command specified\n");

	if (options & (OPT_BIDIR | OPT_SKIP | OPT_PRINTK) && (command != CMD_INSERT))
		 exit_error(2, "-b and -o can only be used with -i (insert)\n");
	if (options & OPT_NUMERIC  &&  command != CMD_LIST)
		 exit_error(2, "-n can only be used with -l (list)\n");
	if ((command == CMD_INSERT) && !(options & OPT_IN || options & OPT_OUT))
		 exit_error(2, "-i can only be used with -I or -O\n");
	if (options & OPT_IN  &&  options & OPT_OUT)
		 exit_error(2, "-I and -O can't be used together\n");

	if ((options & OPT_SOURCE) && (command & (CMD_LIST | CMD_FLUSH )))
		exit_error(2, "no source address (-S) allowed with this command");
	else if (!(options & OPT_SOURCE) && (command & (CMD_INSERT | CMD_DELETE)))
		shostnetworkmask = "0.0.0.0/0";
	if (shostnetworkmask) {
		parse_hostnetworkmask(shostnetworkmask, &addrs, &(nat_set.nat.src_msk), &naddrs);
		nat_set.nat.src = addrs->s_addr;
		parse_all_ports(sports, &(nat_set.nat.ports[0]), (int) nat_set.nat.nsp, nat_set.nat.flag & IP_NAT_SRNG);
	}

	if ((options & OPT_DESTINATION) && (command & (CMD_LIST | CMD_FLUSH)))
		exit_error(2, "no destination address (-D) allowed with this command");
	else if (!(options & OPT_DESTINATION) && (command & (CMD_INSERT | CMD_DELETE)))
		dhostnetworkmask = "0.0.0.0/0";
	if (dhostnetworkmask) {
		parse_hostnetworkmask(dhostnetworkmask, &addrs, &(nat_set.nat.dst_msk), &naddrs);
		nat_set.nat.dst = addrs->s_addr;
		parse_all_ports(dports, &(nat_set.nat.ports[nat_set.nat.nsp]), (int) nat_set.nat.ndp, nat_set.nat.flag & IP_NAT_DRNG);
	}

	if (options & OPT_SRC_NATADDR) {
		parse_hostnetworkmask(new_shostnetworkmask, &addrs, &(nat_set.nat.src_nat_msk), &naddrs);
		nat_set.nat.src_nat = addrs->s_addr;
	}
	if (options & OPT_DST_NATADDR) {
		parse_hostnetworkmask(new_dhostnetworkmask, &addrs, &(nat_set.nat.dst_nat_msk), &naddrs);
		nat_set.nat.dst_nat = addrs->s_addr;
	}


	switch (command) {
		case CMD_DELETE:
			nat_set.command = IP_NAT_CMD_DELETE;
			ret = do_setsockopt(IP_NAT_CNTL,&nat_set,sizeof(struct ip_nat_set));
			break;
		case CMD_INSERT:
			nat_set.command = IP_NAT_CMD_INSERT;
			if (options & OPT_VERBOSE)
				ret = print_rule(&nat_set);
			ret |= do_setsockopt(IP_NAT_CNTL,&nat_set,sizeof(struct ip_nat_set));
			break;
		case CMD_APPEND:
			nat_set.command = IP_NAT_CMD_APPEND;
			if (options & OPT_VERBOSE)
				ret = print_rule(&nat_set);
			ret |= do_setsockopt(IP_NAT_CNTL,&nat_set,sizeof(struct ip_nat_set));
		case CMD_LIST:
			ret = list_nat_rules();
			break;
		case CMD_FLUSH:
			nat_set.command = IP_NAT_CMD_FLUSH;
			ret = do_setsockopt(IP_NAT_CNTL,&nat_set,sizeof(struct ip_nat_set));
			break;
		default:
			exit_tryhelp(2);
	}

	exit(ret);
}



void exit_error(int status, char *msg)
{
	fprintf(stderr, "ipnatadm: %s\n", msg);
	exit_tryhelp(status);
}

void exit_tryhelp(int status)
{
	fprintf(stderr, "Try `ipnatadm -h' for more information.\n");
	exit(status);
}

void exit_printhelp()
{
	printf("%s\n\n"
	       "Usage:	ipnatadm -I command [options] (input NAT)\n"
	       "	ipnatadm -O command [options] (output NAT)\n"
	       "	ipnatadm -F flush all rules\n"
	       "	ipnatadm -h (print this help information))\n\n",
	       package_version);

	printf("Commands:\n"
	       "  -i 	insert rule\n"
	       "  -a    append to rule\n"
	       "  -d 	delete rule\n"
	       "  -l	list all rules of this category\n\n"
	       "Options:\n"
	       "  -Y id (number)\n"
	       "  -P protocol (either tcp, udp, icmp, or all)\n"
	       "  -S address[/mask] [port ...]  (source specification)\n"
	       "  -D address[/mask] [port ...]  (destination specification)\n"
	       "  -M address[/mask] [port] (new src-address and/or port)\n"
	       "  -N address[/mask] [port] (new dst-address and/or port)\n"
	       "  -V address (network interface address)\n"
	       "  -W name    (network interface name)\n"
	       "  -X [roro|byte|conn] (virtual server rule)\n"
	       "  -w weight (number)\n"
	       "  -t timeout (in seconds)\n"
	       "  -b	bidirectional match\n"
	       "  -s	skip NAT (don't translate matching packets)\n"
	       "  -n	numeric output of addresses and ports\n"
	       "  -o	turn on kernel logging for matching packets\n"
	       "  -v	verbose mode\n");

	exit(0);
}
