/*
 *      ipnatadm -- IP NAT administration
 *	(routines.c - helper functions for ipnatadm)
 *
 *      Copyright (c) 1997 by Michael Hasenstein
 *
 *      --------------------------------------------------------------------
 *      large parts of ipnatadm have been taken from ipfwadm, for
 *      this code the following copyright must not be forgotten:
 *
 *            Copyright (c) 1995,1996 by X/OS Experts in Open Systems BV.
 *              All rights reserved.
 *
 *            Author: Jos Vos <jos@xos.nl>
 *
 *              X/OS Experts in Open Systems BV
 *              Kruislaan 419
 *              NL-1098 VA  Amsterdam
 *              The Netherlands
 *
 *              E-mail: info@xos.nl
 *              WWW:    http://www.xos.nl/
 *      --------------------------------------------------------------------
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include "ipnatadm.h"

/*
 * helper routines for ipnatadm
 * they are all from ipfwadm
 */


long string_to_number(char *s, long min, long max)
{
	long number;
	char *end;

	number = strtol(s, &end, 10);
	if (*end == '\0' && end != s) {
		/* we parsed a number, let's see if we want this */
		if (min <= number && number <= max)
			return number;
		else
			return -1;
	} else
		return -1;
}

struct in_addr * parse_hostnetwork(char *name, int *naddrs)
{
	struct in_addr *addrp, *addrptmp;

	if ((addrptmp = dotted_to_addr(name)) != NULL) {
		addrp = nat_malloc(sizeof(struct in_addr));
		inaddrcpy(addrp, addrptmp);
		*naddrs = 1;
		return addrp;
	} else if ((addrptmp = network_to_addr(name)) != NULL) {
		addrp = nat_malloc(sizeof(struct in_addr));
		inaddrcpy(addrp, addrptmp);
		*naddrs = 1;
		return addrp;
	} else if ((addrp = host_to_addr(name, naddrs)) != NULL) {
		return addrp;
	} else {
		fprintf(stderr, "ipnatadm: host/network \"%s\" not found\n", name);
		exit_tryhelp(2);
		/* make the compiler happy... */
		return NULL;
	}
}

int parse_timeout(char *s)
{
	int timeout;

	if ((timeout = string_to_number(s, 0, INT_MAX)) != -1)
		return timeout;
	else {
		fprintf(stderr, "ipnatadm: invalid timeout value \"%s\" specified\n", s);
		exit_tryhelp(2);
		/* make the compiler happy... */
		return 0;
	}
}

void parse_protocol(char *s)
{
	if (strncmp("all", s, strlen(s)) == 0)
		nat_set.nat.protocol = 0;
	else if (strncmp("tcp", s, strlen(s)) == 0)
		nat_set.nat.protocol = IPPROTO_TCP;
	else if (strncmp("udp", s, strlen(s)) == 0)
		nat_set.nat.protocol = IPPROTO_UDP;
	else if (strncmp("icmp", s, strlen(s)) == 0)
		nat_set.nat.protocol = IPPROTO_ICMP;
	else {
		fprintf(stderr, "ipnatadm: invalid protocol \"%s\" specified\n", s);
		exit_tryhelp(2);
		/* make the compiler happy... */
		return;
	}
}

void parse_hostnetworkmask(char *name, struct in_addr **addrpp, struct in_addr *maskp, int *naddrs)
{
	struct in_addr *addrp;
	char buf[256];
	char *p;
	int i, j, k, n;

	strncpy(buf, name, sizeof(buf) - 1);
	if ((p = strrchr(buf, '/')) != NULL) {
		*p = '\0';
		addrp = parse_mask(p + 1);
	} else
		addrp = parse_mask(NULL);
	inaddrcpy(maskp, addrp);

	/* if a null mask is given, the name is ignored, like in "any/0" */
	if (maskp->s_addr == 0L)
		strcpy(buf, "0.0.0.0");

	addrp = *addrpp = parse_hostnetwork(buf, naddrs);
	n = *naddrs;
	for (i = 0, j = 0; i < n; i++) {
		addrp[j++].s_addr &= maskp->s_addr;
		for (k = 0; k < j - 1; k++) {
			if (addrp[k].s_addr == addrp[j - 1].s_addr) {
				(*naddrs)--;
				j--;
				break;
			}
		}
	}
}

void parse_viahost(char *name)
{
	struct in_addr *addrp;
	int naddrs;

	if ((addrp = dotted_to_addr(name)) != NULL) {
		inaddrcpy(&nat_set.nat.via, addrp);
		return;
	} else if ((addrp = host_to_addr(name, &naddrs)) != NULL) {
		if (naddrs != 1) {
			fprintf(stderr,
				"ipnatadm: hostname \"%s\" does not specify a unique address\n", name);
			exit_tryhelp(2);
		} else {
			inaddrcpy(&nat_set.nat.via, addrp);
			return;
		}
	} else {
		fprintf(stderr, "ipnatadm: host \"%s\" not found\n", name);
		exit_tryhelp(2);
	}
}

struct in_addr * parse_mask(char *mask)
{
	static struct in_addr maskaddr;
	struct in_addr *addrp;
	int bits;

	if (mask == NULL) {
		/* no mask at all defaults to 32 bits */
		maskaddr.s_addr = 0xFFFFFFFF;
		return &maskaddr;
	} else if ((addrp = dotted_to_addr(mask)) != NULL) {
		/* dotted_to_addr already returns a network byte order addr */
		return addrp;
	} else if ((bits = string_to_number(mask, 0, 32)) == -1) {
		fprintf(stderr, "ipnatadm: invalid mask \"%s\" specified\n", mask);
		exit_tryhelp(2);
		/* make the compiler happy... */
		return NULL;
	} else if (bits != 0) {
		maskaddr.s_addr = htonl(0xFFFFFFFF << (32 - bits));
		return &maskaddr;
	} else {
		maskaddr.s_addr = 0L;
		return &maskaddr;
	}
}

void parse_all_ports(char **ports, unsigned short *portnumbers, int nports, int range)
{
	int i, j;
	char buf[256], *cp;

	for (i = 0, j = (range ? 2 : 0); i < nports; i++) {
		if (ports[i] == NULL)
			continue;
		strncpy(buf, ports[i], sizeof(buf) - 1);
		if ((cp = strchr(buf, (int) ':')) == NULL)
			portnumbers[j++] = parse_port(buf);
		else {
			*cp = '\0';
			portnumbers[0] = parse_port(buf);
			portnumbers[1] = parse_port(cp + 1);
			if (portnumbers[0] > portnumbers[1]) {
				fprintf(stderr, "ipnatadm: invalid range of ports (%u > %u)\n",
					portnumbers[0], portnumbers[1]);
				exit_tryhelp(2);
			}
		}
	}
}

unsigned short parse_port(char *port)
{
	int portnum;

	if ((portnum = string_to_number(port, 0, 65535)) != -1)
		return (unsigned short) portnum;
	else if (nat_set.nat.protocol == IPPROTO_ICMP) {
		/* ICMP types (given as port numbers) must be numeric! */
		fprintf(stderr, "ipnatadm: invalid ICMP type \"%s\" specified\n", port);
		exit_tryhelp(2);
		/* make the compiler happy... */
		return 0;
	} else if ((portnum = service_to_port(port)) != -1)
		return (unsigned short) portnum;
	else {
		fprintf(stderr, "ipnatadm: invalid port/service \"%s\" specified\n", port);
		exit_tryhelp(2);
		/* make the compiler happy... */
		return 0;
	}
}

void store_port(char *port, unsigned short *nports, int rangeflag, char *ports[])
{
	/* to count the # ports, check whether this is a range or not */
	if (strchr(port, (int) ':') == NULL)
		*nports += 1;
	else if (nat_set.nat.flag & rangeflag)
		exit_error(2, "multiple ranges not allowed");
	else {
		*nports += 2;
		nat_set.nat.flag |= rangeflag;
	}

	if (nat_set.nat.nsp + nat_set.nat.ndp > IP_NAT_MAX_PORTS) {
		fprintf(stderr, "ipnatadm: too many ports specified (maximum %d)\n", IP_NAT_MAX_PORTS);
		exit_tryhelp(2);
	}
	ports[*nports - 1] = port;
	return;
}

struct in_addr * network_to_addr(char *name)
{
	struct netent *net;
	static struct in_addr addr;

	if ((net = getnetbyname(name)) != NULL) {
		if (net->n_addrtype != AF_INET)
			return (struct in_addr *) NULL;
		addr.s_addr = htonl((unsigned long) net->n_net);
		return &addr;
	} else
		return (struct in_addr *) NULL;
}

char * mask_to_dotted(struct in_addr *mask)
{
	int i;
	static char buf[20];
	__u32 maskaddr, bits;

	maskaddr = ntohl(mask->s_addr);

	if (maskaddr == 0xFFFFFFFFL)
		/* we don't want to see "/32" */
		return "";
	else {
		i = 32;
		bits = 0xFFFFFFFEL;
		while (--i >= 0 && maskaddr != bits)
			bits <<= 1;
		if (i >= 0)
			sprintf(buf, "/%d", i);
		else
			/* mask was not a decent combination of 1's and 0's */
			sprintf(buf, "/%s", addr_to_dotted(mask));
		return buf;
	}
}

char * addr_to_dotted(struct in_addr *addrp)
{
	static char buf[20];
	unsigned char *bytep;

	bytep = (unsigned char *) &(addrp->s_addr);
	sprintf(buf, "%d.%d.%d.%d", bytep[0], bytep[1], bytep[2], bytep[3]);
	return buf;
}

struct in_addr * dotted_to_addr(char *dotted)
{
	static struct in_addr addr;
	unsigned char *addrp;
	char *p, *q;
	int onebyte, i;
	char buf[20];

	/* copy dotted string, because we need to modify it */
	strncpy(buf, dotted, sizeof(buf) - 1);
	addrp = (unsigned char *) &(addr.s_addr);

	p = buf;
	for (i = 0; i < 3; i++) {
		if ((q = strchr(p, '.')) == NULL)
			return (struct in_addr *) NULL;
		else {
			*q = '\0';
			if ((onebyte = string_to_number(p, 0, 255)) == -1)
				return (struct in_addr *) NULL;
			else
				addrp[i] = (unsigned char) onebyte;
		}
		p = q + 1;
	}

	/* we've checked 3 bytes, now we check the last one */
	if ((onebyte = string_to_number(p, 0, 255)) == -1)
		return (struct in_addr *) NULL;
	else
		addrp[3] = (unsigned char) onebyte;
		
	return &addr;
}

struct in_addr * host_to_addr(char *name, int *naddr)
{
	struct hostent *host;
	struct in_addr *addr;
	int i;

	*naddr = 0;
	if ((host = gethostbyname(name)) != NULL) {
		if (host->h_addrtype != AF_INET ||
				host->h_length != sizeof(struct in_addr))
			return (struct in_addr *) NULL;
		while (host->h_addr_list[*naddr] != (char *) NULL)
			(*naddr)++;
		addr = nat_calloc(*naddr, sizeof(struct in_addr));
		for (i = 0; i < *naddr; i++)
			inaddrcpy(&(addr[i]), (struct in_addr *) host->h_addr_list[i]);
		return addr;
	} else
		return (struct in_addr *) NULL;
}

int service_to_port(char *name)
{
	struct servent *service;

	if (nat_set.nat.protocol == IPPROTO_TCP && (service = getservbyname(name, "tcp")) != NULL)
		return ntohs((unsigned short) service->s_port);
	else if (nat_set.nat.protocol == IPPROTO_UDP && (service = getservbyname(name, "udp")) != NULL)
		return ntohs((unsigned short) service->s_port);
	else
		return -1;
}

char * port_to_service(int port)
{
	struct servent *service;

	if (nat_set.nat.protocol == IPPROTO_TCP && (service = getservbyport(htons(port), "tcp")) != NULL)
		return service->s_name;
	else if (nat_set.nat.protocol == IPPROTO_UDP && \
		 (service = getservbyport(htons(port), "udp")) != NULL)
		return service->s_name;
	else
		return (char *) NULL;
}

void check_option(long option, char name)
{
	if (options & option) {
		fprintf(stderr, "ipnatadm: multiple -%c flags not allowed\n", name);
		exit_tryhelp(2);
	}
	options |= option;
}

void inaddrcpy(struct in_addr *dst, struct in_addr *src)
{
	/* memcpy(dst, src, sizeof(struct in_addr)); */
	dst->s_addr = src->s_addr;
}

void * nat_calloc(size_t count, size_t size)
{
	void *p;

	if ((p = calloc(count, size)) == NULL) {
		perror("ipnatadm: calloc failed");
		exit(1);
	} else
		return p;
}

int do_setsockopt(int cmd, void *data, int length)
{
	static int sockfd = -1;
	int ret;

	if (sockfd == -1) {
		if ((sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) == -1) {
			perror("ipnatadm: socket creation failed");
			exit(1);
		}
	}
	
	ret = setsockopt(sockfd, IPPROTO_IP, cmd, (char *) data, length);
	if (ret)
		perror("ipnatadm: setsockopt failed");

	return ret;
}

void * nat_malloc(size_t size)
{
	void *p;

	if ((p = malloc(size)) == NULL) {
		perror("ipnatadm: malloc failed");
		exit(1);
	} else
		return p;
}


/* for option '-v' (doesn't list rules created earlier!) */
int print_rule(struct ip_nat_set *nat_set)
{
	int i = 0, ret = 0;

	printf("\n");

	/* print id */
	if (options & OPT_ID)
		printf("id: %d\n", nat_set->nat.id);

	/* print source and destination ip/port(s), new NAT ip/port */
	/* source: */
	printf("src-ip: %s%s", addr_to_dotted(&(nat_set->nat.src)), \
	       mask_to_dotted(&(nat_set->nat.src_msk)));
	if (nat_set->nat.nsp) {
		printf(" (ports(%d): ",nat_set->nat.nsp);
		if (nat_set->nat.flag & IP_NAT_SRNG) {
			printf("%d-%d ",nat_set->nat.ports[0],nat_set->nat.ports[1]);
			i = 2;
		}
		for (; i < nat_set->nat.nsp; i++)
			printf("%d ",nat_set->nat.ports[i]);
		printf("\b)");
	}
	if (options & OPT_SRC_NATADDR) {
		printf(" -> %s%s", addr_to_dotted(&(nat_set->nat.src_nat)), \
		       mask_to_dotted(&(nat_set->nat.src_nat_msk)));
		if (nat_set->nat.src_nat_port) {
			printf(" (%d)", nat_set->nat.src_nat_port);
		}
	}
	printf("\n");
	/* destination: */
	printf("dst-ip: %s%s", addr_to_dotted(&(nat_set->nat.dst)), \
	       mask_to_dotted(&(nat_set->nat.dst_msk)));
	if (nat_set->nat.ndp) {
		i = nat_set->nat.nsp;
		printf(" (ports(%d): ",nat_set->nat.ndp);
		if (nat_set->nat.flag & IP_NAT_DRNG) {
			printf("%d-%d ",nat_set->nat.ports[nat_set->nat.nsp],nat_set->nat.ports[nat_set->nat.nsp+1]);
			i += 2;
		}
		for (; i < nat_set->nat.nsp + nat_set->nat.ndp; i++)
			printf("%d ",nat_set->nat.ports[i]);
		printf("\b)");
	}
		if (options & OPT_DST_NATADDR) {
		printf(" -> %s%s", addr_to_dotted(&(nat_set->nat.dst_nat)), \
		       mask_to_dotted(&(nat_set->nat.dst_nat_msk)));
		if (nat_set->nat.dst_nat_port) {
			printf(" (%d)", nat_set->nat.dst_nat_port);
		}
	}
	printf("\n");

	/* print protocol */
	printf("protocol: ");
	if (nat_set->nat.protocol == 0)
		printf("all\n");
	else if (nat_set->nat.protocol == IPPROTO_TCP)
		printf("tcp\n");
	else if (nat_set->nat.protocol == IPPROTO_UDP)
		printf("udp\n");
	else if (nat_set->nat.protocol == IPPROTO_ICMP)
		printf("icmp\n");

	/* print interface ip */
	if (options & OPT_VIAADDR)
		printf("interface ip: %s\n", addr_to_dotted(&(nat_set->nat.via)));

	/* print interface name */
	if (options & OPT_VIANAME)
		printf("interface name: %s\n", nat_set->vianame);

	/* print flags */
	printf("options:");
	if (nat_set->nat.flag & IP_NAT_IN)
		printf(" \"incoming\"");
	if (nat_set->nat.flag & IP_NAT_OUT)
		printf(" \"outgoing\"");
	if (nat_set->nat.flag & IP_NAT_BIDIR)
		printf(" \"bidirectional\"");
	if (nat_set->nat.flag & IP_NAT_SKIP)
		printf(" \"skip\"");
	if (nat_set->nat.flag & IP_NAT_LOG)
		printf(" \"log packets\"");
	printf("\n");
	
	return ret;
}

int list_nat_rules()
{
	int ret = 0;
	return ret;
}
